/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file enclosure.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _ENCLOSURE_HPP
#define _ENCLOSURE_HPP

#include <string>
#include <stdexcept>
#include "dynsys/cntaylintf.h"
#include "dynsys/taylorex.h"
//#include "interval/doubleFun.h"

namespace capd{
namespace dynsys{

template<typename MapType>
typename CnTaylor<MapType>::VectorType CnTaylor<MapType>::enclosure(const VectorType &x, int *found) const
// the function finds an enclosure for \varphi([0,step],x)
{
   ScalarType trialStep = ScalarType(-0.2,1.2)*m_step;
   int dimension = getDimension();
   VectorType y(dimension),z(dimension),Small(dimension);

   ScalarType h = ScalarType(0,1)*m_step;

   double multf = 1.5;
   *found=0;
   int counter=0,limit=10,i;

   for(i=0;i<dimension;++i)
      Small[i] = ScalarType(-1,1)*1e-15;

   z = x + trialStep * m_vectorField(x);
   z += Small; // to be sure that z has noempty interior


   while((!(*found)) && (counter<limit))
   {
      counter++;
      y = x + h * m_vectorField(z);
      *found = 1;
      for(i=0;i<dimension;++i)
      {
         if(!(y[i].subsetInterior(z[i])))
         {
            *found = 0;
            z[i] = y[i];
            ScalarType s;
            z[i].split(s);
            s = multf*s;
            z[i] += s;
         }
      }
   }

   if(!(*found))
      throw TaylorException<VectorType>("CnTaylor error: cannot find enclosure guaranteeing bounds",x,m_step);

   return y;
}

//###########################################################//


template<typename MapType>
typename CnTaylor<MapType>::MatrixType CnTaylor<MapType>::jacEnclosure(const VectorType &enc, NormType &the_norm) const
// the function finds enclosure for Jacobian matrix (variational part)
// source- "C^1-Lohner algorithm" by P. Zgliczynski
{
   int dimension = getDimension();
   MatrixType der = m_vectorField[enc], result(dimension,dimension);

   m_l = the_norm(der).rightBound(); // computation of lagarithmic norm
   ScalarType s = ScalarType(0,1)*m_step;
   ScalarType w = ScalarType(-1,1)*exp(s*m_l);

   MatrixType W(dimension,dimension); // W_3 in paper "C^1 - Lohner algorithm"
   W = w;

   result = MatrixType::Identity(dimension) + s*der*W;

   int i,j;
   for(i=1;i<=dimension;++i)
      for(j=1;j<=dimension;++j)
      {
         ScalarType d = result(i,j);
         typename ScalarType::BoundType 
              l = (w.leftBound() > d.leftBound() ? w.leftBound() : d.leftBound()),
              r = (w.rightBound() < d.rightBound() ? w.rightBound() : d.rightBound());
         result(i,j) = ScalarType(l,r);
      }

   return result;
}

//###########################################################//

template<typename MapType>
void CnTaylor<MapType>::c2Enclosure(
         const VectorType &W1, // rough enclosure for x
         const MatrixType &W3, // rough enclosure for V
         NormType &norm,       // a logarithmic norm
         C2CoeffType& result
      ) const
{
   int i,j,c,s,r;
   MatrixType DF = m_vectorField[W1];
   EuclNorm<VectorType,MatrixType> tempNorm;
   ScalarType l = norm(DF).rightBound();    // l>= max x\in\W_1 \mu(\frac{df}{dx})

   int dimension = getDimension();
   C2CoeffType temp(dimension);
   m_vectorField.computeSecondDerivatives(W1,temp);

   for(j=0;j<dimension;++j)
   {
      for(c=j;c<dimension;++c)
      {
         VectorType Bjc(dimension);
         Bjc.clear();
         for(i=0;i<dimension;++i)
         {
            for(s=0;s<dimension;++s)
               for(r=0;r<dimension;++r)
                  Bjc[i] += temp(i,s,r) * W3[r][c] * W3[s][j];
         }
         ScalarType delta = tempNorm(Bjc).rightBound();
         double size;
         if(l.contains(0.0))
            size = abs(delta*abs(m_step).rightBound()).rightBound();
         else
            size = (abs(delta * (exp(l*m_step)-ScalarType(1.))/l)).rightBound();
         for(i=0;i<dimension;++i)
            result(i,j,c) = ScalarType(-size,size);
      } // c - loop
   } // j - loop
}

//###########################################################//

template<typename MapType>
typename CnTaylor<MapType>::VectorType CnTaylor<MapType>::cnEnclosure(const VectorType& x, NormType& the_norm, int rank) const
{
   int found,i,j;
   VectorType enc = enclosure(x,&found);
   MatrixType jacEnc = jacEnclosure(enc,the_norm);
   m_vectorField.computeDerivatives(enc,m_vectorFieldSeries,0,0,rank);
   setInitialCondition(enc,jacEnc);

// computation of rough enclosure for C^d part, d=2,...,rank
   for(i=2;i<=rank;++i)
   {
      computeNonlinearPart(i);
      Multipointer mp = m_resultSeries.first(i);
      do{
         ScalarType delta = computeDelta(mp).rightBound();
         double size;
         if(subset(0.0,m_l))
            size = abs(delta*abs(m_step).rightBound()).rightBound();
         else
            size = (abs(delta * (exp(m_l*m_step)-ScalarType(1.))/m_l)).rightBound();
         ScalarType res(-size,size);
         for(j=0;j<getDimension();++j)
            m_resultSeries(j,mp)->value[0] = res;
      }while(m_resultSeries.next(mp));
   }
   return enc;
}

//###########################################################//

template<typename MapType>
void CnTaylor<MapType>::cnEnclosure(const VectorType& x, CnCoeffType& coeff, NormType& the_norm) const
{
   cnEnclosure(x,the_norm,coeff.rank());
   for(int i=0;i<getDimension();++i)
   {
      typename SeriesContainerType::const_iterator b = m_resultSeries.begin(i), e = m_resultSeries.end(i,coeff.rank());
      typename CnCoeffType::iterator j = coeff.begin(i);
      while(b!=e)
      {
         (*j) = (*b)->value[0];
         ++b;
         ++j;
      }
   }
}

//###########################################################//

template<typename MapType>
typename CnTaylor<MapType>::ScalarType CnTaylor<MapType>::computeDelta(const Multipointer& mp) const
{
   VectorType v(getDimension());
   for(int i=0;i<getDimension();++i)
   {
      v[i] = ( m_resultSeries(i,mp)->value[0] ) = m_nonlinearPart(i,mp)->operator()(0);
   }
   EuclNorm<VectorType,MatrixType> tempNorm;
   return tempNorm(v);
}

}} // namespace capd::dynsys

#endif // #define _ENCLOSURE_HPP
/// @}
