/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file plusminus.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

/** Author: Daniel Wilczak, 2004-2005 */

#ifndef _PLUSMINUS_HPP
#define _PLUSMINUS_HPP

namespace capd{
namespace map{

// ---------------------- constructors --------------------------------------------

template<typename ScalarType>
SumNode<ScalarType>::SumNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

template<typename ScalarType>
DifNode<ScalarType>::DifNode(Node<ScalarType> *a_left, Node<ScalarType> *a_right, int a_order)
   : Node<ScalarType>(a_order,a_left,a_right)
{
   value = new ScalarType[m_order];
   iterator b=begin(), e=end();
   while(b!=e)
   {
      *b = ScalarType(0);
      ++b;
   }
   if(left) ++(left->m_links);
   if(right) ++(right->m_links);
}

// --------------------------- destructors ---------------------------------------

template<typename ScalarType>
SumNode<ScalarType>::~SumNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

template<typename ScalarType>
DifNode<ScalarType>::~DifNode()
{
   if( !( --(left->m_links)) ) 
      delete left;
   if( !( --(right->m_links)) ) 
      delete right;
   delete []value;
}

// -------------------------- operators ------------------------------------------

template<typename ScalarType>
ScalarType& SumNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   return value[i] = (*left)(i) + (*right)(i);
}

template<typename ScalarType>
ScalarType& DifNode<ScalarType>::operator()(int i)
{
   if(i<=m_maxComputedDerivative)
      return value[i];
   m_maxComputedDerivative = i;

   return value[i] = (*left)(i) - (*right)(i);
}

// -------------------------- setOrder ------------------------------------------

template<typename ScalarType>
void SumNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];

   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

template<typename ScalarType>
void DifNode<ScalarType>::setOrder(int a_order, ScalarType *a_val)
{
   if(m_order==a_order) return;
   delete [] value;

   m_order = a_order;
   value = new ScalarType[m_order];

   if(left) left->setOrder(m_order,a_val);
   if(right) right->setOrder(m_order,a_val);
}

}} // namespace capd::map

// --------------------------------- global Operators -------------------------

template<typename ScalarType>
capd::map::Node<ScalarType>& operator+(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y)
{
   if(x.getOrder()!=y.getOrder())
      throw std::runtime_error("operator+(Node&, Node&) - incompatible dimensions");
   return *(new capd::map::SumNode<ScalarType>(&x,&y,x.getOrder()));
}

template<typename ScalarType>
capd::map::Node<ScalarType>& operator-(capd::map::Node<ScalarType>& x, capd::map::Node<ScalarType>& y)
{
   if(x.getOrder()!=y.getOrder())
      throw std::runtime_error("operator-(Node&, Node&) - incompatible dimensions");
   return *(new capd::map::DifNode<ScalarType>(&x,&y,x.getOrder()));   
}

#endif // define _PLUSMINUS_HPP

/// @}
