/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file ivector.hpp
///
/// @author The CAPD Group
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _I_VECTOR_HPP_
#define _I_VECTOR_HPP_

#include <stdexcept>
#include "vectalg/iobject.hpp"

//---------------  definitions for class intervalVector only -------------------//

template<typename IVectorType>
IVectorType diam(const IVectorType& v)
{
   IVectorType result(v.dimension());
   diameter(v,result); // defined in iobject.hpp
   return result;
}

/******************************************************************************/
// intersection of two interval vectors
// halts if the intersection is empty
/******************************************************************************/

template<typename IVectorType>
IVectorType intersection(const IVectorType &v1, const IVectorType &v2)
{
   IVectorType result(v1.dimension());
   intersection(v1,v2,result); // defined in iobject.hpp
   return result;
}

template<typename IVectorType>
IVectorType midVector(const IVectorType& v)
{
   IVectorType result(v.dimension());
   mid(v,result); // defined in iobject.hpp
   return result;
}

template<typename IVectorType>
IVectorType intervalHull(const IVectorType& v1, const IVectorType& v2)
{
   IVectorType result(v1.dimension());
   intervalHull(v1,v2,result);
   return result;
}

template<typename IVectorType>
IVectorType leftVector(const IVectorType& v)
{
   IVectorType temp(v.dimension());
   typename IVectorType::const_iterator b=v.begin(), e=v.end();
   typename IVectorType::iterator i = temp.begin();
   while(b!=e)
   {
      *i = b->leftBound();
      ++i;
      ++b;
   }
   return temp;
}

template<typename IVectorType>
IVectorType rightVector(const IVectorType& v)
{
   IVectorType temp(v.dimension());
   typename IVectorType::const_iterator b=v.begin(), e=v.end();
   typename IVectorType::iterator i = temp.begin();
   while(b!=e)
   {
      *i = b->rightBound();
      ++i;
      ++b;
   }
   return temp;
}

template<typename IVectorType>
IVectorType intervalBall(const IVectorType &iv, const typename IVectorType::ScalarType &r)
{
   IVectorType result(iv.dimension());
   typename IVectorType::iterator b=result.begin(), e=result.end();
   typename IVectorType::const_iterator i=iv.begin();

   while(b!=e)
   {
      *b = ball(*i,r);
      ++b;
      ++i;
   }
   return result;
}

template<typename IVectorType>
typename IVectorType::ScalarType solveAffineInclusion(
      const IVectorType& a,
      const IVectorType& p,
      const IVectorType& c
   )
{
   typedef typename IVectorType::ScalarType ScalarType;

   if(a.dimension()!=p.dimension() || a.dimension()!=c.dimension())
      throw std::runtime_error("Incompatible vectors in function solveAffineInclusion");

   ScalarType result = solveAffineInclusion(a[0],p[0],c[0]);
   for(int i=1; i<a.dimension(); ++i)
   {
      ScalarType iv=solveAffineInclusion(a[i],p[i],c[i]);
      result = capd::min( result.leftBound(), iv.leftBound() );
   }
   return result;
}

template<typename IVectorType>
typename IVectorType::ScalarType solveAffineInclusion(
      const IVectorType& a,
      const IVectorType& p,
      const IVectorType& c,
      int& dir
   )
{
   if(a.dimension()!=p.dimension() || a.dimension()!=c.dimension())
      throw std::runtime_error("Incompatible vectors in function solveAffineInclusion");

   typedef typename IVectorType::ScalarType ScalarType;
   ScalarType result = solveAffineInclusion(a[0],p[0],c[0]);
   dir=0;

   for(int i=1; i<a.dimension(); ++i)
   {
      ScalarType iv = solveAffineInclusion(a[i],p[i],c[i]);
      if( iv.leftBound() < result.leftBound() )
      {
         result = iv.leftBound();
         dir = i;
      }
   }
   return result;
}

#endif // _I_VECTOR_HPP_
/// @}
