/// @addtogroup vectalg
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file matrixContainer.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file is part of the CAPD library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

#ifndef _MatrixContainer_H
#define _MatrixContainer_H

#include "vectalg/container.h"

namespace capd{
namespace vectalg{

template<typename Scalar, int rows, int cols>
class MatrixContainer : public Container<Scalar,rows*cols>
{
public:
   typedef Scalar ScalarType;
   typedef typename Container<Scalar,rows*cols>::iterator iterator;
   typedef typename Container<Scalar,rows*cols>::const_iterator const_iterator;
   typedef MatrixContainer ContainerType;
   typedef Container<Scalar,rows> ColumnContainer;
   typedef Container<Scalar,cols> RowContainer;

   inline MatrixContainer() : Container<Scalar,rows*cols>() {}
   inline MatrixContainer(int a_rows, int a_cols) : Container<Scalar,rows*cols>(a_rows*a_cols){}
   inline MatrixContainer(const MatrixContainer& mc) : Container<Scalar,rows*cols>(mc) {}
   inline MatrixContainer(int a_rows, int a_cols,bool) : Container<Scalar,rows*cols>(a_rows*a_cols,true){}

   inline int numberOfRows() const {return rows;}
   inline int numberOfColumns() const {return cols;}

   inline MatrixContainer& operator=(const MatrixContainer& a)
   {
      Container<Scalar,rows*cols>::operator=(a);
      return *this;
   }
   inline static void* operator new[](size_t sizeOfData)
   {
      return ::operator new[](sizeOfData);
   }
   inline static void* operator new[](size_t sizeOfData, int, int)
   {
      return ::operator new[](sizeOfData);
   }

   void assertEqualDimension(const MatrixContainer& c) const
   {
      if(rows!=c.numberOfRows() || cols!=c.numberOfColumns())
         throw std::runtime_error("Unequal dimensions of containers in function: MatrixContainer::assertEqualDimensions");
   }
   
   using Container<Scalar,rows*cols>::begin;
   using Container<Scalar,rows*cols>::end;
   using Container<Scalar,rows*cols>::dimension;

protected:
   inline void resize(int r, int c)
   {
      if(r!=rows || c!=cols)
      {
         throw std::range_error("Cannot resize static MatrixContainer");
      }
   }
   using Container<Scalar,rows*cols>::data;
};

// ---------------------------------------------------------------------

template<typename Scalar>
class MatrixContainer<Scalar,0,0> : public Container<Scalar,0>
{
public:
   typedef Scalar ScalarType;
   typedef typename Container<Scalar,0>::iterator iterator;
   typedef typename Container<Scalar,0>::const_iterator const_iterator;
   typedef MatrixContainer ContainerType;
   typedef Container<Scalar,0> ColumnContainer;
   typedef Container<Scalar,0> RowContainer;

   inline MatrixContainer()
      : Container<Scalar,0>(defaultRows*defaultCols),
        m_rows(defaultRows), m_cols(defaultCols)
   {}
   inline MatrixContainer(int a_rows, int a_cols)
      : Container<Scalar,0>(a_rows*a_cols), m_rows(a_rows), m_cols(a_cols)
   {}
   inline MatrixContainer(const MatrixContainer& mc)
      : Container<Scalar,0>(mc), m_rows(mc.m_rows), m_cols(mc.m_cols)
   {}
   inline MatrixContainer(int a_rows, int a_cols,bool)
      : Container<Scalar,0>(a_rows*a_cols,true), m_rows(a_rows), m_cols(a_cols)
   {}
   friend void swap(MatrixContainer<Scalar,0,0>& A_m1, MatrixContainer<Scalar,0,0>& A_m2)
   {
       std::swap(*static_cast<Container<Scalar,0>*>(&A_m1),*static_cast<Container<Scalar,0>*>(&A_m2));
       std::swap(A_m1.m_rows,A_m2.m_rows);
       std::swap(A_m1.m_cols,A_m2.m_cols);
   }

   inline int numberOfRows() const {return m_rows;}
   inline int numberOfColumns() const {return m_cols;}

   MatrixContainer& operator=(const MatrixContainer& a)
   {
      Container<Scalar,0>::operator=(a);
      m_rows = a.m_rows;
      m_cols = a.m_cols;
      return *this;
   }
   static void* operator new[](size_t sizeOfData)
   {
      return ::operator new[](sizeOfData);
   }
   static void* operator new[](size_t sizeOfData, int defR, int defC)
   {
      defaultRows = defR;
      defaultCols = defC;
      return ::operator new[](sizeOfData);
   }

   //using Container<Scalar,0>::operator[];
   using Container<Scalar,0>::begin;
   using Container<Scalar,0>::end;
   using Container<Scalar,0>::dimension;

protected:
   int m_rows,m_cols;
   static int defaultRows, defaultCols;

   using Container<Scalar,0>::resize;
   inline void resize(int r, int c)
   {
      if(r!=m_rows || c!=m_cols)
      {
         Container<Scalar,0>::resize(r*c);
         m_rows = r;
         m_cols = c;
      }
   }
   using Container<Scalar,0>::data;
};

template<typename Scalar>
int MatrixContainer<Scalar,0,0>::defaultRows=0;

template<typename Scalar>
int MatrixContainer<Scalar,0,0>::defaultCols=0;

}} // namespace capd::vectalg

#endif

/// @}
