// this file contains rigorous routines used in the proof of the existence 
// of a branch of period two points connecting first and second period doubling bifurcations
// this file in fact implements Algorithm 1 from the paper
// which uses C^3-computations to prove the existence of continuation of period two branch

// ----------------------------------------------------------------------------
// auxiliary routine
// given a MainData structure which describes coordinate system and set 
// we create the next (larger) set for continuation of period two points 

void initNextSet(const MainData& md, MainData& o_nextSet, 
   double nextA, const IVector& box,
   double d3gStep, int d3gOrder, int d3gGridX, int d3gGridA,
   const char d3gEstimation[]
)
{
   o_nextSet = md;
   o_nextSet.mina = md.maxa;
   o_nextSet.maxa = nextA;
   
   o_nextSet.box = box;
      
   o_nextSet.d3gStep  = d3gStep;
   o_nextSet.d3gOrder = d3gOrder;
   o_nextSet.d3gGridX = d3gGridX;
   o_nextSet.d3gGridA = d3gGridA;
   o_nextSet.d2gd3gEstimations = d3gEstimation;
}


// ----------------------------------------------------------------------------
// part of Algorithm 1. Given a set described in MainData md we verify if 
// dG/da > 0 on left edge
// dG/da < 0 on right edge
// this, together with condition verified in function checkSignsMaxA
// G(maxa,left)<0 and G(maxa,right)>0
// implies that
// G([mina,maxa],left)<0 and G([mina,maxa],right)>0

template<typename PoincareMapType, typename MainDataType>
void computeLeftRightEdges(PoincareMap& PM, PoincareMapType& PM4D, const MainDataType& md,int order, double step, int grid)
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename LInterval::BoundType BType;
   
   PM.setStep(LInterval(step));
   PM.setOrder(order);
   PM4D.setStep(LInterval(step));
   PM4D.setOrder(order);
   
// we divide range of parameter values onto grid equal parts
   LInterval skipA = (LInterval(md.maxa)-LInterval(md.mina))/grid;
   LInterval boxA = LInterval(0.,1.)*skipA;

// foreach subinterval in grid we proceed
   for(int i=0;i<grid;i++)
   {
      LInterval a = LInterval(md.mina) + i*skipA + boxA;
      std::cout << "\na: " << a << std::endl;
      
      // compute derivative dG/da(a,left)
      LInterval y1 = findYOfX(PM,a,md.box[0].left(),md);
      LInterval dgda_y1 = computeDGDA(PM4D,md.box[0].left(),y1,a,md);
      std::cout << "dgda left: " << dgda_y1 << std::endl;
      if(!(dgda_y1>0))
         throw std::runtime_error("dgda left is not positive");

      // compute derivative dG/da(a,right)
      LInterval y2 = findYOfX(PM,a,md.box[0].right(),md);
      LInterval dgda_y2 = computeDGDA(PM4D,md.box[0].right(),y2,a,md);
      std::cout << "dgda right: " << dgda_y2 << std::endl;
      if(!(dgda_y2<0))
         throw std::runtime_error("dgda right is not negative");
   }
}

// ----------------------------------------------------------------------------
// this function verifies that for all a\in[mina,maxa] there exists unique fixed point in the set 
// given the estimation for the fixed point the procedure verifies if there exists a point on the left side of fixed point
// for which G is positive
// and on the right side for which G is negative. This implies that G has at least three zeros for each parameter value

template<typename PoincareMapType, typename MainDataType>
void computeFixedPointsBranch(
         PoincareMapType& PM, 
         PoincareMapType& PM4D, 
         const MainDataType& md,
         int order, double step, int grid,double t
      )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename LInterval::BoundType BType;
   
   PM.setStep(LInterval(step));
   PM.setOrder(order);
   PM4D.setStep(LInterval(step));
   PM4D.setOrder(order);
   
// divide range of parameters onto grid parts
   LInterval skipA = (LInterval(md.maxa)-LInterval(md.mina))/grid;
   LInterval boxA = LInterval(0.,1.)*skipA;

   for(int i=0;i<grid;i++)
   {
      LInterval a = LInterval(md.mina) + i*skipA + boxA;
      std::cout << "\na: " << a << std::endl;

      PM.setParameter("a",a);
      // for a fixed subinterval verify the existence of unique fixed point
      LVector Nx = provePeriodicOrbit(PM,md.Ix,md.ICoordSystem,md.IInvCS,md.box,2*md.bifNumber);

      std::cout << "Nx(A): " << Nx << std::endl;

      // next we choose a point between left edge and fixed point
      LInterval x1 = md.box[0].left()*(1-t)+t*Nx[0].left();
      LInterval y1 = findYOfX(PM,a,x1,md);

      // this block verifies that G(a,chosenPoint) is positive
      // first verify that G(mina,chosenPoint)>0
      LInterval Gx1 = computeG(PM,a.left(),x1,md);
      std::cout << Gx1 << std::endl;
      if( !(Gx1>0) )
      {
         throw std::runtime_error("intermediate point left: G is not positive");
      }
      // and that DG/da(a,chosenPoint)>0
      
      LInterval dgda_y1 = computeDGDA(PM4D,x1,y1,a,md);
      std::cout << "dgda left: " << dgda_y1 << std::endl;

      if(!(dgda_y1>0))
         throw std::runtime_error("dgda left is not positive");

      // the same for some point between fixed point and right edge

      LInterval x2 = md.box[0].right()*(1-t)+t*Nx[0].right();
      LInterval y2 = findYOfX(PM,a,x2,md);
      LInterval Gx2 = computeG(PM,a.left(),x2,md);
      std::cout << Gx2 << std::endl;
      if( !(Gx2<0) )
      {
         throw std::runtime_error("intermediate point right: G is not negative");
      }

      LInterval dgda_y2 = computeDGDA(PM4D,x2,y2,a,md);
      std::cout << "dgda right: " << dgda_y2 << std::endl;

      if(!(dgda_y2<0))
         throw std::runtime_error("dgda right is not negative");
   }
}

// ----------------------------------------------------------------------------
// this procedure just calls Algorithm 1 five times with arguments md0, md1, ...
void proveConnectingBranch(BasicPoincareMap& BPM,PoincareMap& PM, PoincareMap& PM4D, MainData& md)
{
   IVector box(2);
   box[0] = interval(-9.9022e-3,9.795e-3);
   box[1] = interval(-4e-4,4e-4);

   MainData md0(md);
   initNextSet(md,md0,
      2.8325/*nextA*/,
      box,
      0.04/*d3gStep*/, 
      5/*d3gOrder*/, 
      150/*d3gGridX*/, 
      8/*d3gGridA*/,
      "md0_d3g.dat"
   );

   setPMSettings(PM,0.06,14);
   // verify if G(maxa,left)<0 and G(maxa,right)>0
   // and verify if dG/da([mina,maxa],left)>0, dG/da([mina,maxa],right)<0.
   // Both conditions implies G([mina,maxa],left)<0, G([mina,maxa],right)>0
   checkSignsMaxA(PM,md0);
   computeLeftRightEdges(PM,PM4D,md0,10,0.03,70);

   // this function verifies that for all a\in[mina,maxa] there exists unique fixed point in the set 
   // given the estimation for the fixed point the procedure verifies if there exists a point on the left side of fixed point
   // for which G is positive
   // and on the right side for which G is negative. This implies that G has at least three zeros for each parameter value
   computeFixedPointsBranch(PM,PM4D,md0,10,0.03,600,.88);

   // finally we verify that D3G/dx >0 which implies that G has exactly three zeros for each parameter
   checkD3G(PM,md0,false);

// -----------------------------------------

   box[0]=interval(-1.669e-2,1.639e-2);
   box[1]=interval(-4e-4,4e-4);
   MainData md1(md0);
   initNextSet(md0,md1,
      2.8326/*nextA*/,
      box,
      0.04/*d3gStep*/,
      6/*d3gOrder*/,
      270/*d3gGridX*/,
      14/*d3gGridA*/,
      "md1_d3g.dat"
   );

   setPMSettings(PM,0.05,14);
   checkSignsMaxA(PM,md1);
   computeLeftRightEdges(PM,PM4D,md1,10,0.03,80);
   computeFixedPointsBranch(PM,PM4D,md1,10,0.03,300,0.7);
   checkD3G(PM,md1,false);

// -----------------------------------------

   box[0]=interval(-2.1459e-2,2.0962e-2);
   box[1]=interval(-4e-4,4e-4);
   MainData md2(md1);
   initNextSet(md1,md2,
      2.8327/*nextA*/,
      box,
      0.04/*d3gStep*/, 
      6/*d3gOrder*/, 
      340/*d3gGridX*/, 
      16/*d3gGridA*/,
      "md2_d3g.dat"
   );

   setPMSettings(PM,0.05,14);
   checkSignsMaxA(PM,md2);
   computeLeftRightEdges(PM,PM4D,md2,10,0.03,60);
   computeFixedPointsBranch(PM,PM4D,md2,10,0.03,150,0.6);
   checkD3G(PM,md2,false);

// -----------------------------------------

   box[0]=interval(-2.538e-2,2.468e-2);
   box[1]=interval(-8e-4,8e-4);
   MainData md3(md2);
   initNextSet(md2,md3,
      2.8328/*nextA*/,
      box,
      0.04/*d3gStep*/, 
      6/*d3gOrder*/, 
      340/*d3gGridX*/, 
      17/*d3gGridA*/,
      "md3_d3g.dat"
   );

   setPMSettings(PM,0.05,14);
   checkSignsMaxA(PM,md3);
   computeLeftRightEdges(PM,PM4D,md3,10,0.03,50);
   computeFixedPointsBranch(PM,PM4D,md3,10,0.03,120,0.5);
   checkD3G(PM,md3,false);

// -----------------------------------------

   box[0]=interval(-2.87757e-2,2.79e-2);
   box[1]=interval(-8e-4,8e-4);
   MainData md4(md3);
   initNextSet(md3,md4,
      2.8329/*nextA*/,
      box,
      0.04/*d3gStep*/, 
      6/*d3gOrder*/, 
      350/*d3gGridX*/, 
      20/*d3gGridA*/,
      "md4_d3g.dat"
   );

   setPMSettings(PM,0.05,14);
   checkSignsMaxA(PM,md4);
   computeLeftRightEdges(PM,PM4D,md4,10,0.03,50);
   computeFixedPointsBranch(PM,PM4D,md4,10,0.03,100,0.5);
   checkD3G(PM,md4,false);
}
