// ----------------------------------------------------------------------------
// this is rigorous routine
// it computes value of function G(a,x) = x - P^2(a,x,y(a,x))
// where y(a,x) is given from Liapunov-Schmidt reduction
// arguments: 
// a - parameter of a system
// x - initial condition
// md - the structure which defines in particular coordinate system (different for first and second period doubling)
template<typename PoincareMapType, typename MainDataType>
typename PoincareMapType::ScalarType computeG(
      PoincareMapType& PM, 
      typename PoincareMapType::ScalarType a, 
      typename PoincareMapType::ScalarType x, 
      MainDataType& md
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

//  here we compute an upper bound for y(a,x)
   LInterval y = findYOfX(PM,a,x,md);
   LVector u(2);
   u[0]=x; u[1]=y;
   LVector center = midVector(u);
   LVector box = u-center;
   // compute a Poincare map in euclidean coordinate system
   LVector image = computePoincareMap(PM,md.Ix + md.ICoordSystem*center,md.ICoordSystem,box,4*md.bifNumber);
   // change the coordinates
   image = changeCoordSystem(md.Ix,md.IInvCS,image);
   return x - image[0];
}

// ----------------------------------------------------------------------------
// this is rigorous routine
// it computes value of function G(a,x) = x - P^2(a,x,y(a,x))
// and its derivative \partial G/ \partial x

template<typename PoincareMapType, typename MainDataType>
typename PoincareMapType::ScalarType computeG( 
      PoincareMapType& PM, 
      typename PoincareMapType::ScalarType a, 
      const typename PoincareMapType::VectorType& u, 
      MainDataType& md, 
      typename PoincareMapType::ScalarType& o_dGx
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;
   typedef typename PoincareMapType::MatrixType LMatrix;
   typedef typename PoincareMapType::CnCoeffType LCnCoeff;

   LVector center = midVector(u);
   LVector box = u-center;
   
   LCnCoeff DP2(2,1);
   LVector image = computePoincareMap(PM,md.Ix + md.ICoordSystem*center,md.ICoordSystem,box,4*md.bifNumber,DP2);
   image = changeCoordSystem(md.Ix,md.ICoordSystem,md.IInvCS,image,DP2);

// here we use derivative of implicit function y(x)
   LInterval dydx = DP2(1,0)/(1.-DP2(1,1));
   o_dGx = 1 - DP2(0,0) - DP2(0,1)*dydx;

   return u[0] - image[0];
}

// ----------------------------------------------------------------------------
// this is rigorous routine
// it computes value of function G(a,x) = x - P^2(a,x,y(a,x))
// and its derivative \partial G/ \partial x

template<typename PoincareMapType, typename MainDataType>
typename PoincareMapType::ScalarType computeG( 
      PoincareMapType& PM, 
      typename PoincareMapType::ScalarType a, 
      typename PoincareMapType::ScalarType x, 
      MainDataType& md, 
      typename PoincareMapType::ScalarType& o_dGx,
      bool findBetterEnclosure=true
   )
{
   typedef typename PoincareMapType::ScalarType LInterval;
   typedef typename PoincareMapType::VectorType LVector;

   LInterval y = findYOfX(PM,a,x,md,findBetterEnclosure);
   LVector u(2);
   u[0]=x; u[1]=y;
   return computeG(PM,a,u,md,o_dGx);
}
