// this file contains all the procedures which verify the existence of PDB on some set stored 
// in data structure MainData md

// ----------------------------------------------------------------------------

// this function computes that between bifurcation the eigenvalue of derivative of Poincare Map
// at the fixed point is larger than -1
// it returns computed upper bound for the fixed point
IVector checkEigenvalueMinA(BasicPoincareMap& BPM, PoincareMap& PM, const MainData& md)
{
   std::cout << "\n\n----------------- Eigenvalue MinA -------------\n";
   PM.setStep(md.minaStep);
   PM.setOrder(md.minaOrder);
   BPM.setParameter("a",md.mina);
   PM.setParameter("a",md.mina);
   std::cout << "mina=" << md.mina << std::endl;

   interval ev1,ev2;
   IVector x = computeEigenvalues(BPM,PM,md,ev1,ev2,true);
   if(!(ev1>-1.))
      throw std::runtime_error("Eigenvalue at a=mina is not grater than -1");

   return x;
}

// ----------------------------------------------------------------------------
// similar procedure for eigenvalues after bifurcation
// here we use additionaly mpmd data - for computations in MultiPrecision arithmetics

MPVector checkEigenvalueMaxA(BasicPoincareMap& BPM, MPPoincareMap& PM, const MPMainData& mpmd)
{
   std::cout << "\n\n----------------- Eigenvalue MaxA -------------\n";

   BPM.setParameter("a",mpmd.maxa);
   PM.setParameter("a",MPInterval(mpmd.maxa));
   std::cout << "maxa=" << mpmd.maxa << std::endl;

   MPInterval ev1,ev2;
   MPVector x = computeEigenvalues(BPM,PM,mpmd,ev1,ev2,true);
   if(!(ev1<-1.))
      throw std::runtime_error("Eigenvalue at a=maxa is not smaler than -1");

   return x;
}


// ----------------------------------------------------------------------------
// this function verifies if after bifurcation we have
// G(maxa,left)  < 0
// G(maxa,right) > 0
template<typename PoincareMapType, typename MainDataType>
void checkSignsMaxA(PoincareMapType& PM, MainDataType& md)
{
   typedef typename PoincareMapType::ScalarType LInterval;

   LInterval Gx1 = computeG(PM,LInterval(md.maxa),md.box[0].left(),md);
   std::cout << Gx1 << std::endl;
   if( !(Gx1<0) )
   {
      throw std::runtime_error("checkMaxA: inclusion not satisfied at maxX");
   }

   LInterval Gx2 = computeG(PM,LInterval(md.maxa),md.box[0].right(),md);
   std::cout << Gx2 << std::endl;
   if( !(Gx2>0) )
   {
      throw std::runtime_error("checkMaxA: inclusion not satisfied at maxX");
   }   
}

// ----------------------------------------------------------------------------
// this procedure calls the previous two with suitable parameters
// moreover, given an estimation for the fixed point after bifurcation
// we verify if dG/dx(maxa,fp)<0
// this is one of the assumptions of the main bifurcation theorem
void checkMinMaxA(BasicPoincareMap& BPM, PoincareMap& PM, MainData& md)
{
   std::string rosslerFormula="var:x,y,z,a,b;fun:-(y+z),x+b*y,b+z*(x-a);";
   std::string section = "var:x,y,z;fun:x;";
   capd::multiPrec::MpfrClass::setDefaultPrecision(150);

   MPPoincareMap MPPM(rosslerFormula,section,md.maxaOrder,MPInterval(md.maxaStep));
   MPPM.setParameter("b",MPInterval(2)/MPInterval(10));
   MPMainData mpmd(md);

   checkSignsMaxA(MPPM,mpmd);
   std::cout.precision(20);

   checkEigenvalueMinA(BPM,PM,md);
   MPVector N = checkEigenvalueMaxA(BPM,MPPM,mpmd); 

   MPInterval dGx;
   computeG(MPPM,MPInterval(md.maxa),N[0],mpmd,dGx);
   std::cout << "dGx(maxa,xfp(maxa)) = " << dGx << std::endl;
   if(!(dGx<0))
      throw std::runtime_error("checkDGX: dGx(maxa,xfp(maxa)) is not negative");   

}
// ----------------------------------------------------------------------------

// this is rigorous routine
// it proves the existence of a fixed point for P^2 for all parameter values from a given range
// moreover, it returns a bound for x_{fp}'([mina,maxa])

interval computeDXA(PoincareMap* PM4D[], PoincareMap* PM[], const MainData& md)
{
   setPMSettings(PM,md.dxa3DStep,md.dxa3DOrder);
   setPMSettings(PM4D,md.dxa4DStep,md.dxa4DOrder);

   interval skipA = (interval(md.maxa)-interval(md.mina))/md.dxaGridA;
   interval boxA = interval(0,1)*skipA;

   interval *DXDA = new interval[omp_get_max_threads()];
   bool *first = new bool[omp_get_max_threads()];
   for(int j=0;j<omp_get_max_threads();++j) first[j] = true;

   int i=0;
   #pragma omp  parallel default(none) shared(md,PM, PM4D, skipA, std::cout, boxA,first,DXDA) private(i)
   {
      #pragma omp for nowait
      for(i=0;i<md.dxaGridA;i++)
      {
         interval a = interval(md.mina) + i*skipA + boxA;
         PM[omp_get_thread_num()]->setParameter("a",a);
         IVector Nx = provePeriodicOrbit(*PM[omp_get_thread_num()],md.Ix,md.ICoordSystem,md.IInvCS,md.box,2*md.bifNumber);
         std::cout << "Nx(A): " << Nx << std::endl;
         interval dxda =  computeDXA(*PM4D[omp_get_thread_num()],Nx[0],Nx[1],a,md);
         if(first[omp_get_thread_num()])
         {
            DXDA[omp_get_thread_num()] = dxda;
            first[omp_get_thread_num()] = false;
         } else
            DXDA[omp_get_thread_num()] = intervalHull(DXDA[omp_get_thread_num()],dxda);
      }
   }
   interval result = DXDA[0];
   for(int k=1;k<omp_get_max_threads();++k)
   {
      // if first[k]==true then this thread was not used at all
      if(!first[k])
         result = intervalHull(result,DXDA[k]);
   }
   delete []first;
   delete []DXDA;
   std::cout << "dxda(A): " << result << std::endl;
   return result;
}

// ----------------------------------------------------------------------------
// this function verifies that the derivative dG/dx(mina,[minx,maxx])>0
// which means that before bifurcation there exists at most one fixed point
// the obtained bound is write to file
void checkDGX(PoincareMap* PM[], const MainData& md)
{
   setPMSettings(PM,md.dgxStep,md.dgxOrder);

   interval *bound = new interval[omp_get_max_threads()];
   bool *first = new bool[omp_get_max_threads()];
   for(int j=0;j<omp_get_max_threads();++j) first[j] = true;

   int i=0;
   #pragma omp  parallel default(none) shared(md,PM, std::cout,first,bound) private(i)
   {
      #pragma omp for nowait
      for(i=0;i<(int)md.dgxGridX.size();i++)
      {
         interval x = md.dgxGridX[i];
         interval dGx;
         computeG(*PM[omp_get_thread_num()],md.mina,x,md,dGx,false);
         std::cout << "step: " << i << " of " << md.dgxGridX.size() << ". dGx(mina,x) = " << dGx << std::endl;
         if(first[omp_get_thread_num()])
         {
            bound[omp_get_thread_num()] = dGx;
            first[omp_get_thread_num()] = false;
         } else
            bound[omp_get_thread_num()] = intervalHull(dGx,bound[omp_get_thread_num()]);
         if(!(bound[omp_get_thread_num()]>0))
            throw std::runtime_error("checkDGX: dGx(mina,x) is not positive on [x1,x2]");
      }
   }
   interval result = bound[0];
   for(int k=1;k<omp_get_max_threads();++k)
   {
      // if first[k]==true then this thread was not used at all
      if(!first[k])
          result = intervalHull(result,bound[k]);
   }
   delete []first;
   delete []bound;

   std::ofstream aFile;
   aFile.open(md.dgdxOnA1.c_str());
   aFile.precision(20);
   aFile << "DG_DX:" << result << std::endl;
   aFile.close();
}


// ----------------------------------------------------------------------------
// this function verifies, if D3G >0  on Z = [mina,maxa] \times [minx,maxx]
// moreover, it returns D2G on Z

interval checkD3G(PoincareMap* PM[], const MainData &md, bool findBetterEnclosure=false)
{
   interval D3G;
   interval D2G;

   setPMSettings(PM,md.d3gStep,md.d3gOrder);
   std::cout << "-------------------------------------\n";
   std::cout << "The proof that G'''(x)>0\n";

   // first the set is divided onto d3gGridA \times d3gGridX parts

   interval stepA = (interval(md.maxa)-interval(md.mina))/md.d3gGridA;
   interval stepX = interval(diam(md.box[0]))/md.d3gGridX;
   interval boxA = interval(0,1)*stepA;
   interval boxX = interval(0,1)*stepX;

   interval *D3Gbound = new interval[omp_get_max_threads()];
   interval *D2Gbound = new interval[omp_get_max_threads()];
   bool *first = new bool[omp_get_max_threads()];
   for(int j=0;j<omp_get_max_threads();++j) first[j] = true;

   int i=0;
   #pragma omp  parallel default(none) shared(md,PM, std::cout,first,D3Gbound,D2Gbound, findBetterEnclosure, boxX, stepX, stepA, boxA) private(i)
   {
      #pragma omp for nowait
      // next for each part we proceed as follows
      for(i=0;i<md.d3gGridA;i++)
      {
         for(int j=0;j<md.d3gGridX;j++)
         {
            interval d2g,d3g;
            // here we compute both d^3G/dx^3 and d^2G/dx^2
            computeD3G( *PM[omp_get_thread_num()],
                        md.mina+i*stepA+boxA,
                        md.box[0].leftBound()+j*stepX+boxX,
                        md, d2g, d3g,
                        findBetterEnclosure
                     );
            // the variables D3G and D2G contain bounds for d^3G/dx^3 and d^2G/dx^2 on the whole set
            if(first[omp_get_thread_num()])
            {
               D2Gbound[omp_get_thread_num()] = d2g;
               D3Gbound[omp_get_thread_num()] = d3g;
               first[omp_get_thread_num()] = false;
            } else
            {
               D2Gbound[omp_get_thread_num()] = intervalHull(D2Gbound[omp_get_thread_num()],d2g);
               D3Gbound[omp_get_thread_num()] = intervalHull(D3Gbound[omp_get_thread_num()],d3g);
            }
            std::cout << "step " << (i+1) << "x" << (j+1) << " of " << md.d3gGridA << "x" << md.d3gGridX << ". ";
            std::cout << "G'''(x) = " << d3g << std::endl;
            // D3G should be nonzero. We verify, if it is positive
            if(!(D3Gbound[omp_get_thread_num()]>0))
               throw std::runtime_error("checkD3G: D3G is not positive");
         }
      }
   }

   // intervalHull of results from different threads
   D3G = D3Gbound[0];
   D2G = D2Gbound[0];
   for(int k=1;k<omp_get_max_threads();++k)
   {
      // if first[k]==true then this thread was not used at all
      if(!first[k])
      {
         D3G = intervalHull(D3G,D3Gbound[k]);
         D2G = intervalHull(D2G,D2Gbound[k]);
      }
   }
   delete [] D3Gbound;
   delete [] D2Gbound;
   delete [] first;

   // finally, the procedure stores estimations in files
   std::ofstream aFile;
   aFile.open(md.d2gd3gEstimations.c_str());
   aFile.precision(20);

   aFile << "D3G:";
   writeInterval(aFile,D3G);
   aFile << std::endl;

   aFile << "D2G:";
   writeInterval(aFile,D2G);
   aFile << std::endl;

   aFile.close();

   // the function returns computed d^2G/dx^2 which is necessary for verifying other conditions
   return D2G;
}


// ----------------------------------------------------------------------------

// this function verifies if the function x_fp(a)
void checkMonotonicity(PoincareMap* PM4D[], PoincareMap* PM[], const MainData& md)
{
   interval D2GDX2 = checkD3G(PM,md);
   interval dx_fp = computeDXA(PM4D,PM,md);

   setPMSettings(PM,md.d2g3DStep,md.d2g3DOrder);
   setPMSettings(PM4D,md.d2g4DStep,md.d2g4DOrder);

   interval stepA = (interval(md.maxa)-interval(md.mina))/md.d2gGridA;
   interval stepX = interval(diam(md.box[0]))/md.d2gGridX;
   interval boxA = interval(0,1)*stepA;
   interval boxX = interval(0,1)*stepX;

   interval *D2GDXDAbound = new interval[omp_get_max_threads()];
   bool *first = new bool[omp_get_max_threads()];
   for(int j=0;j<omp_get_max_threads();++j) first[j] = true;

   for(int i=0;i<md.d2gGridA;i++)
   {
      interval a = md.mina+i*stepA+boxA;
      for(int j=0;j<md.d2gGridX;j++)
      {
         interval x = md.box[0].leftBound()+j*stepX+boxX;
         interval y = findYOfX(*PM[omp_get_thread_num()],a,x,md,false);
         interval d2gdxda = computeD2G_DXDA(*PM4D[omp_get_thread_num()],x,y,a,md);
         if(first[omp_get_thread_num()])
         {
            D2GDXDAbound[omp_get_thread_num()] = d2gdxda;
            first[omp_get_thread_num()] = false;
         } else
            D2GDXDAbound[omp_get_thread_num()] = intervalHull(D2GDXDAbound[omp_get_thread_num()],d2gdxda);

         std::cout << d2gdxda << std::endl;
      }
      std::cout << "step " << i << " of " << md.d2gGridA << " done" << std::endl;   
      std::cout << "z: " << (D2GDXDAbound[omp_get_thread_num()] + D2GDX2*dx_fp*interval(0,1)) << std::endl;
   }

   // intervalHull of results from different threads
   interval D2GDXDA = D2GDXDAbound[0];
   for(int k=1;k<omp_get_max_threads();++k)
   {
      // if first[k]==true then this thread was not used at all
      if(!first[k])
      {
         D2GDXDA = intervalHull(D2GDXDA,D2GDXDAbound[k]);
      }
   }
   delete [] first;
   delete [] D2GDXDAbound;

   interval z = D2GDXDA + D2GDX2*dx_fp*interval(0,1);

   std::ofstream aFile;
   aFile.open(md.d2gdxaEstimation.c_str());
   aFile.precision(20);

   aFile << "D2G_DXDA:";
   writeInterval(aFile,D2GDXDA);
   aFile << std::endl;

   aFile << "D2GDXDA + D2GDX2*dx_fp*interval(0,1):";
   writeInterval(aFile,z);
   aFile << std::endl;

   aFile.close();

   std::cout << z << std::endl;
   if(!(z<0))
      throw std::runtime_error("checkMonotonicity: inclusion not verified");
}


// ----------------------------------------------------------------------------

void computeEigenvalues(PoincareMap& PM, PoincareMap& PM4D,const MainData& md)
{
   PM.setStep(md.ev3DStep);
   PM.setOrder(md.ev3DOrder);
   PM4D.setStep(md.ev4DStep);
   PM4D.setOrder(md.ev4DOrder);

   interval skipA = (interval(md.maxa)-interval(md.mina))/md.evGridA;
   interval boxA = interval(0,1)*skipA;

   interval EV1, EV2, DEV1;
   IMatrix DP;
   bool first=true;

   for(int i=0;i<md.evGridA;i++)
   {
      interval a = interval(md.mina) + i*skipA + boxA;
      interval ev1, ev2;
      IMatrix dp;
      interval dev1 = computeEigenvalues(PM,PM4D,md,a,ev1,ev2,dp);
      std::cout << "step: " << (i+1) << " of " << md.evGridA << std::endl;
      std::cout << "ev1: " << ev1 << std::endl;
      std::cout << "ev2: " << ev2 << std::endl << std::endl;
      std::cout << "d\\lambda_1/da = " << dev1 << std::endl << std::endl;

      if(dev1.contains(0))
         throw std::runtime_error("derivative of eigenvalue containts zero!");
      if(!(capd::abs(ev2)<1))
         throw std::runtime_error("second eigenvalue is not smaler than 1!");
      if(first)
      {
         EV1 = ev1;
         EV2 = ev2;
         DEV1 = dev1;
         DP = dp;
         first = false;
      } else
      {
         EV1 = intervalHull(EV1,ev1);
         EV2 = intervalHull(EV2,ev2);
         DEV1 = intervalHull(DEV1,dev1);
         DP = intervalHull(DP,dp);
      }
   }
   std::ofstream aFile;
   aFile.open(md.eigenvalues.c_str());
   aFile.precision(20);
   aFile << "lambda1(A): " << EV1 << std::endl;
   aFile << "lambda2(A): " << EV2 << std::endl;
   aFile << "dlambda1/da(A): " << DEV1 << std::endl;
   aFile << "\\pi_x(eigenvector1):" << md.IInvCS*DP;
   aFile.close();
}

// ----------------------------------------------------------------------------

void checkA4aCondition(PoincareMap& PM, MainData& md)
{
   PM.setStep(md.a4aStep);
   PM.setOrder(md.a4aOrder);

   interval stepA = (interval(md.maxa)-interval(md.mina))/md.a4aGridA;
   interval stepX = interval(diam(md.box[0]))/md.a4aGridX;
   interval boxA = interval(0,1)*stepA;
   interval boxX = interval(0,1)*stepX;
   bool first = true;

   interval yEstimation;
   for(int i=0;i<md.a4aGridA;i++)
   {
      interval a = md.mina+i*stepA+boxA;
      PM.setParameter("a",a);
      for(int j=0;j<md.a4aGridX;j++)
      {
         IVector center(2), box(2);
         center[0] = md.box[0].leftBound()+j*stepX+boxX;
         center[1] = md.box[1];
         split(center,box);
         IVector im = computePoincareMap(PM,md.Ix+md.ICoordSystem*center,md.ICoordSystem,box,2*md.bifNumber);
         im = changeCoordSystem(md.Ix,md.IInvCS,im);
         std::cout << im << std::endl;

         if(first)
         {
            yEstimation = im[1];
            first = false;
         } else
            yEstimation = intervalHull(yEstimation,im[1]);

         if(!(yEstimation.subsetInterior(md.box[1])))
            throw std::runtime_error("A4a condtion not verified!");
      }
      std::cout << "step " << i << " of " << md.a4aGridA << " done" << std::endl;   
   }

   std::ofstream aFile;
   aFile.open(md.a4aEstimation.c_str());
   aFile.precision(20);
   aFile << "P_y(A,Y,Z): " << yEstimation << std::endl;
   aFile.close();
}
