// auxiliary routine
// it sets order and step to each PoincareMap object in the table

template<typename PoincareMapType>
void setPMSettings(PoincareMapType* PM[], interval step, int order)
{
   typedef typename PoincareMapType::ScalarType LInterval;
   for(int i=0;i<omp_get_max_threads();++i)
   {
      PM[i]->setStep(LInterval(step));
      PM[i]->setOrder(order);
   }
}

// ----------------------------------------------------------------------------
// this function computes eigenvalues of a 2x2 matrix
template<typename MatrixType>
void eigenvalues(
      const MatrixType& m, 
      typename MatrixType::ScalarType& ev1, 
      typename MatrixType::ScalarType& ev2
   )
{
   typedef typename MatrixType::ScalarType Scalar;
   Scalar a=m(1,1);
   Scalar b=m(1,2);
   Scalar c=m(2,1);
   Scalar d=m(2,2);
   Scalar s = sqrt((a-d)*(a-d) + 4*b*c);
   ev1 = (a + d - s)/2.;
   ev2 = (a + d + s)/2.;
}

// ----------------------------------------------------------------------------
// this function computes eigenvalues of a 2x2 matrix together with
// derivative of ev1 with respect to the parameter of the system
// argument m is assumed to contain partial derivatives of Poincare map 
// up to the second order both with respect to start point on the section
// and with respect to the parameter

// the formula for dev1 is given in the paper
template<typename CnCoeffType>
void eigenvalues(
      const CnCoeffType& m, 
      typename CnCoeffType::ScalarType& ev1, 
      typename CnCoeffType::ScalarType& ev2,
      typename CnCoeffType::ScalarType& dev1
   )
{
   typedef typename CnCoeffType::ScalarType Scalar;
   Scalar a=m(0,0);
   Scalar b=m(0,1);
   Scalar c=m(1,0);
   Scalar d=m(1,1);
   Scalar s = sqrt((a-d)*(a-d) + 4*b*c);
   ev1 = (a + d - s)/2.;
   ev2 = (a + d + s)/2.;

   Scalar det = (m(0,0)-1)*(m(1,1)-1)-m(0,1)*m(1,0);
   Scalar dxda = ( m(0,2)*(1-m(1,1)) + m(0,1)*m(1,2) )/det;
   Scalar dyda = ( m(1,2)*(1-m(0,0)) + m(1,0)*m(0,2) )/det;

   CnCoeffType cp = m;
   for(int i=0;i<2;++i)
   {
      for(int j=0;j<2;++j)
      {
         cp(i,j,2) = m(i,j,2) + m(i,j,0)*dxda + m(i,j,1)*dyda;
      }
   }
   dev1 =( cp(0,0,2) + cp(1,1,2)
            - (
                  2.*(cp(1,0)*cp(0,1,2) + cp(0,1)*cp(1,0,2))
                  +(cp(0,0)-cp(1,1))*(cp(0,0,2)-cp(1,1,2)) 
              )/s
         )/2.;
}

// ----------------------------------------------------------------------------
// the following function computes normalized eigenvectors 
// of a 2x2 matrix
// the columns of the result matrix correspond to eigenvalues ev1, ev2 computed 
// in previous procedures

template<typename MatrixType>
MatrixType eigenvectors(const MatrixType& m)
{
   typedef typename MatrixType::ScalarType Scalar;
   MatrixType result(2,2);

   Scalar a=m(1,1);
   Scalar b=m(1,2);
   Scalar c=m(2,1);
   Scalar d=m(2,2);

   Scalar s = sqrt(a*a + 4*b*c - 2*a*d + d*d);
   result(1,1) = -(a - d - s);
   result(2,1) = -2*c;
   result(1,2) = a - d + s;
   result(2,2) = 2*c;
   result.column(0).normalize();
   result.column(1).normalize();
   return result;
}


// ----------------------------------------------------------------------------
// direct formula for computing inverse of 2x2 martix
template<typename MatrixType>
MatrixType inverse2x2Matrix(const MatrixType& m)
{
   typedef typename MatrixType::ScalarType Scalar;
   Scalar  det = m(1,1)*m(2,2)-m(1,2)*m(2,1);
   if(isSingular(det)) 
   {
      std::cout << "det: " << det << std::endl;
      throw std::runtime_error("inverse2x2Matrix - cannot inverse singular matrix");
   }
   MatrixType result(2,2);
   result(1,1) = m(2,2)/det;
   result(1,2) = -m(1,2)/det;
   result(2,1) = -m(2,1)/det;
   result(2,2) = m(1,1)/det;
   return result;
}

// ----------------------------------------------------------------------------
// this function returns 2x2 submatrix of 3x3 matrix m
// first row and first column are removed

template<typename MatrixType>
MatrixType take2DMatrix(const MatrixType& m)
{
   MatrixType result(2,2);
   result(1,1) = m(2,2);
   result(1,2) = m(2,3);
   result(2,1) = m(3,2);
   result(2,2) = m(3,3);
   return result;
}

// ----------------------------------------------------------------------------
// this function extends 2x2 matrix to 3x3 matrix
// it returns matrix of the form 
// | 1 0 |
// | 0 m |
template<typename MatrixType>
MatrixType expandTo3DMatrix(const MatrixType& m)
{
   MatrixType result(3,3);
   result(1,1) = m(1,1);
   result(1,2) = m(1,2);
   result(2,1) = m(2,1);
   result(2,2) = m(2,2);
   result(3,3) = 1;
   return result;
}

// ----------------------------------------------------------------------------
// this function returns 2-dim vector from 3-dim vector
// first coefficient is removed
template<typename VectorType>
VectorType take2DVector(const VectorType& v)
{
   VectorType result(2);
   result[0] = v[1];
   result[1] = v[2];
   return result;
}

// ----------------------------------------------------------------------------
// this function creates 3-dim vector from 2-dim vector.
// the first coefficient is set to zero, the other two are from 
// argumetn v
template<typename VectorType>
VectorType expandTo3DVector(const VectorType& v)
{
   VectorType result(3);
   result[0] = v[0];
   result[1] = v[1];
   result[2] = 0;
   return result;
}

// ----------------------------------------------------------------------------
// given a vector 'v' represented in euclidean coordinates
// the function computes its coordinates in affine coordinates (coordSystem,x)
// the function has an argument which is an inverse matrix of the matrix which defines coordinateSystem
template<typename MatrixType>
inline typename MatrixType::RowVectorType changeCoordSystem(
      const typename MatrixType::RowVectorType& x, 
      const MatrixType& invCoordSystem, 
      const typename MatrixType::RowVectorType& v
   )
{
   return invCoordSystem*(v-x);
}

// ----------------------------------------------------------------------------
// given partial derivatives of a Poincare map up to some order (parameter and result 'ao_der')
// the function recomputes these derivatives to other coordinate system given by affine transformation
template<typename CnCoeffType>
inline typename CnCoeffType::VectorType changeCoordSystem(
      const typename CnCoeffType::VectorType& x, 
      const typename CnCoeffType::MatrixType& coordSystem, 
      const typename CnCoeffType::MatrixType& invCoordSystem, 
      const typename CnCoeffType::VectorType& v, 
      CnCoeffType& ao_der
   )
{
   ao_der = capd::normalForms::linearSubstitution(invCoordSystem,ao_der,coordSystem);
   return invCoordSystem*(v-x);
}

// ----------------------------------------------------------------------------
// this function computes interval hull of two matrices
// i.e. a subset of C = R^{n\times n} such that A\subset C and B\subset C
template <typename ScalarT,int r, int c>
Matrix<ScalarT,r,c> intervalHull(const Matrix<ScalarT,r,c>& A, const Matrix<ScalarT,r,c>& B)
{
   Matrix<ScalarT,r,c> result(A.numberOfRows(),A.numberOfColumns());
   if( (A.numberOfRows()!=B.numberOfRows()) || (A.numberOfColumns()!=B.numberOfColumns()) )
      throw std::runtime_error("intervalHull(Matrix,Matrix) - incompatible dimensions.");
   typename Matrix<ScalarT,r,c>::iterator i = result.cbegin();
   typename Matrix<ScalarT,r,c>::const_iterator b1 = A.cbegin(), b2=B.cbegin(), e1=A.cend();
   while(b1!=e1)
   {
      (*i) = intervalHull(*b1,*b2);
      ++i;
      ++b1;
      ++b2;
   }
   return result;
}
